Index: FL/Fl_Image.H
===================================================================
--- FL/Fl_Image.H	(revision 10301)
+++ FL/Fl_Image.H	(working copy)
@@ -76,6 +76,7 @@
   static void labeltype(const Fl_Label *lo, int lx, int ly, int lw, int lh, Fl_Align la);
   static void measure(const Fl_Label *lo, int &lw, int &lh);
 
+
   public:
 
   /** 
@@ -179,8 +180,14 @@
   friend class Fl_Xlib_Graphics_Driver;
   static size_t max_size_;
   static Fl_RGB_Scaling scaling_;
-public:
 
+  protected:
+
+  friend class Fl_Pixmap;
+  static Fl_RGB_Image* copy_bilinear(const Fl_RGB_Image& im, int W, int H);
+
+  public:
+
   const uchar *array;
   int alloc_array; // Non-zero if array was allocated
 
Index: src/Fl_Image.cxx
===================================================================
--- src/Fl_Image.cxx	(revision 10301)
+++ src/Fl_Image.cxx	(working copy)
@@ -67,6 +67,8 @@
   }
 }
 
+#include "rgb_image_copy_bilinear.cxx"
+
 /**
   The copy() method creates a copy of the specified
   image. If the width and height are provided, the image is
@@ -237,6 +239,7 @@
   }
   if (W <= 0 || H <= 0) return 0;
 
+  if (scaling_ == FL_SCALING_NEAREST) {
   // OK, need to resize the image data; allocate memory and 
   uchar		*new_ptr;	// Pointer into new array
   const uchar	*old_ptr;	// Pointer into old array
@@ -261,7 +264,6 @@
   new_image = new Fl_RGB_Image(new_array, W, H, d());
   new_image->alloc_array = 1;
 
-  if (scaling_ == FL_SCALING_NEAREST) {
     // Scale the image using a nearest-neighbor algorithm...
     for (dy = H, sy = 0, yerr = H, new_ptr = new_array; dy > 0; dy --) {
       for (dx = W, xerr = W, old_ptr = array + sy * line_d; dx > 0; dx --) {
@@ -285,66 +287,7 @@
     }
   } else {
     // Bilinear scaling
-    const float xscale = (w() - 1) / (float) W;
-    const float yscale = (h() - 1) / (float) H;
-    for (dy = 0; dy < H; dy++) {
-      float oldy = dy * yscale;
-      if (oldy >= h())
-        oldy = h() - 1;
-      const float yfract = oldy - (unsigned) oldy;
-
-      for (dx = 0; dx < W; dx++) {
-        new_ptr = new_array + dy * W * d() + dx * d();
-
-        float oldx = dx * xscale;
-        if (oldx >= w())
-          oldx = w() - 1;
-        const float xfract = oldx - (unsigned) oldx;
-
-        const unsigned leftx = oldx;
-        const unsigned lefty = oldy;
-        const unsigned rightx = oldx + 1 >= w() ? oldx : oldx + 1;
-        const unsigned righty = oldy;
-        const unsigned dleftx = oldx;
-        const unsigned dlefty = oldy + 1 >= h() ? oldy : oldy + 1;
-        const unsigned drightx = rightx;
-        const unsigned drighty = dlefty;
-
-        uchar left[4], right[4], downleft[4], downright[4];
-        memcpy(left, array + lefty * line_d + leftx * d(), d());
-        memcpy(right, array + righty * line_d + rightx * d(), d());
-        memcpy(downleft, array + dlefty * line_d + dleftx * d(), d());
-        memcpy(downright, array + drighty * line_d + drightx * d(), d());
-
-        int i;
-        if (d() == 4) {
-          for (i = 0; i < 3; i++) {
-            left[i] *= left[3] / 255.0f;
-            right[i] *= right[3] / 255.0f;
-            downleft[i] *= downleft[3] / 255.0f;
-            downright[i] *= downright[3] / 255.0f;
-          }
-        }
-
-	const float leftf = 1 - xfract;
-	const float rightf = xfract;
-	const float upf = 1 - yfract;
-	const float downf = yfract;
-
-        for (i = 0; i < d(); i++) {
-          new_ptr[i] = (left[i] * leftf +
-                   right[i] * rightf) * upf +
-                   (downleft[i] * leftf +
-                   downright[i] * rightf) * downf;
-        }
-
-        if (d() == 4 && new_ptr[3]) {
-          for (i = 0; i < 3; i++) {
-            new_ptr[i] /= new_ptr[3] / 255.0f;
-          }
-        }
-      }
-    }
+	 new_image = copy_bilinear(*this, W, H);
   }
 
   return new_image;
Index: src/Fl_Pixmap.cxx
===================================================================
--- src/Fl_Pixmap.cxx	(revision 10301)
+++ src/Fl_Pixmap.cxx	(working copy)
@@ -298,6 +298,7 @@
   }
   if (W <= 0 || H <= 0) return 0;
 
+  if (Fl_RGB_Image::scaling() == FL_SCALING_NEAREST) {
   // OK, need to resize the image data; allocate memory and
   char		**new_data,	// New array for image data
 		**new_row,	// Pointer to row in image data
@@ -380,6 +381,14 @@
 
   new_image = new Fl_Pixmap((char*const*)new_data);
   new_image->alloc_data = 1;
+  }
+  else {
+    // Bilinear scaling
+    Fl_RGB_Image *im = new Fl_RGB_Image(this); // temp. convert Fl_Pixmap to Fl_RGB_Image
+    Fl_Image *ni = Fl_RGB_Image::copy_bilinear(*im, W, H);
+    delete im;
+    return ni;
+  }
 
   return new_image;
 }
Index: src/rgb_image_copy_bilinear.cxx
===================================================================
--- src/rgb_image_copy_bilinear.cxx	(revision 0)
+++ src/rgb_image_copy_bilinear.cxx	(working copy)
@@ -0,0 +1,385 @@
+// Written (c) by Roman Kantor
+// The code is Free software distributed unfer LGPL license with FLTK exceptions, see http://www.fltk.org/COPYING.php
+
+// TWO-dimensional scalling is performend in two steps by rescalling first in one direction
+// and then in the other - which at the end has the same effect as direct two-dimensional rescalling...
+// All rutines are performed with "integer" types for the speed, precision is not lost because the result is scalled up
+// and only after resampling in both directions the values are normalised (scalled back) to the range 0-255.
+// The macro rutines are written to be general and can be used with different types, number of channels,
+// sample padding (stride), row padding, can encode/decode the channels from the original data etc...
+
+
+// ONE-DIMENSIONAL ROUTINES
+
+// a) DOWNSAMPLING (w2<w1):
+// Following method seems to work better than classic (bi)linear interpolation as all the original pixels contribute to the
+// resulting resampled images with the same weight - that is no pixels are lost or suppressed.
+//  Downsampling is performed by "weighted" pixel combinations:
+// 1) If the whole original pixel lies within the area of a new pixel, its whole value contributes to that particular new pixel.
+// 2) If an original pixel lies on the boundary of two new pixels, its contribution is split to these two pixels with
+//    a ratio corresponding to the areas within these pixels. The sum of these two contributions is equal to the whole
+//    contribution as in case 1)
+// All the contributions are then summed up, the sum of weights gives the normalization (scalling) factor equal to f = w1.
+// This factor should be used later to divide all result values thus scalling them back to its original range
+// (usually to range 0-255 for 8-bit channels for RGB(A) images)
+
+
+// b) UPSAMPLING:
+// For upsampling linear combination is performed where resulting pixel value is  P = a2 * P1 + a1 * P2
+// The contributing factors a2, a1 are proportional to the (swapped) distances from  the surrounding original pixels.
+// The normalizarion factor is equal to f = a2 + a1 = 2 * w2
+
+
+
+// Macro parameters:
+// T_SOURCE: type of source data (eg unsigned char)
+// T_TARGET: type of ratrget tata (eg unsigned int)
+// T_ACU type of intermediate data. This type hould be a type "big enough" to hold the scalled values so it does not overflow after all pixel combination.
+//       The maximum value is equal to maximum pixel value times normalization factor. For instance for downscalling if T_SOURCE type is 8-bit
+//       "unsigned char", T_ACU can be 32-bit "unsigned int" without overflow for all images with original width w1 < 2^24.  However if you combine
+//       row/collumn resampling, the scale factor is multiplication of both and could easily oferflow 32 bitif scalling down images with sizes above 4096x4096.
+//       For that reason we use 64-bit type for T_ACU for the second scalling during second dimension resampling.
+// source - (const) pointer to source data array
+// target - pointer to target data array
+// source_stride - pointer shift between samples (normaly for packed data it is equal to number of channels)
+// target_stride - pinter shift between resulting data
+// PREPROCESS - pre-processing sequence (code) for raw data stored in "input" variable (of type T_ACU). This is useful ie if multi-channels are encoded
+//       within single sample (like when ARGB channels are encoded within single "int" sample), getting red channel can be performed by  putting sequence
+//           val >>=16; val &= 0xFF;
+//       in the place of PREPROCESS parameter.
+// POSTPROCESS_STORE - post-processing sequence (code), can be used eg for rescalling back the data or channel encoding where symbol "output" holds
+//       raw interpolated data and "tg" is the pointer where the result should be stored. This pointer can already have an initial value.
+//       For instance for multi-channel encoding and "red channel" you can write something like
+//           *tg |= (((output + (scale_factor>>1)/scale_factor) & 0xFF) << 16;
+//        If you do not perform additional rescalling (*tg = output;) the value in "output" are scalled by the normalization factor
+//       (f = w1    for downscalling and    f = 2 * w2)
+
+#define RESAMPLE_DOWN(T_SOURCE, T_TARGET, T_ACU, w1, w2, source, target, source_stride, target_stride, PREPROCESS, POSTPROCESS_STORE) \
+  { \
+    const T_SOURCE * src = source; \
+    T_TARGET * tg = target; \
+    T_ACU pre = 0; \
+    T_ACU input; \
+    T_ACU output; \
+    unsigned rest = w1; \
+    for(unsigned int i = w2 - 1; i; i--) { \
+      T_ACU full = 0; \
+      while(rest>=w2) { \
+        input = *src; \
+        PREPROCESS \
+        src += source_stride; \
+        full += input; \
+        rest -= w2; \
+      } \
+      input = *src; \
+      PREPROCESS \
+      src += source_stride; \
+      output = pre + w2 * full + rest * input; \
+      POSTPROCESS_STORE \
+      tg += target_stride; \
+      pre = (w2 - rest) * input; \
+      rest += w1 - w2; \
+    } \
+    output = 0; \
+    while(rest>=w2) { \
+      input = *src;\
+      PREPROCESS \
+      output += input;\
+      src += source_stride; \
+      rest -= w2; \
+    } \
+    output *= w2; \
+    output += pre; \
+    POSTPROCESS_STORE \
+  }
+
+// Resampling by linear interpolation
+// This is similar to RESAMPLE_DOWN however this time the result is scalled by factor  f = 2 * w2  - unless modified by POSTPROCESS_STORE sequence.
+#define RESAMPLE_UP(T_SOURCE, T_TARGET, T_ACU, w1, w2, source, target, source_stride, target_stride, PREPROCESS, POSTPROCESS_STORE) \
+  { \
+    unsigned  W2 = w2 * 2; \
+    unsigned W1 = w1 * 2; \
+    const T_SOURCE * src = source; \
+    T_TARGET * tg = target; \
+    T_ACU input = *src; \
+    PREPROCESS \
+    T_ACU output = input * W2; \
+    T_ACU tmp = output; \
+    unsigned i = 0; \
+    unsigned mod = w1; \
+    while(mod<=w2){ \
+      POSTPROCESS_STORE \
+      output = tmp; \
+      mod += W1; \
+      tg += target_stride; \
+      i++; \
+    } \
+    mod -= w2; \
+    T_ACU pre_input = input; \
+    src += source_stride; \
+    input = *src; \
+    PREPROCESS \
+    for(unsigned no = w2 - 2 * i; no; no--) { \
+      if(mod>=W2){ \
+        mod -=W2; \
+        pre_input = input; \
+        src += source_stride;\
+        input = *src; \
+        PREPROCESS \
+      } \
+      T_ACU output = input * mod + (W2 - mod) * pre_input; \
+      POSTPROCESS_STORE \
+      tg += target_stride; \
+      mod += W1;\
+    } \
+    output = W2 * input; \
+    tmp = output; \
+    while(i){\
+      POSTPROCESS_STORE \
+      tg += target_stride; \
+      output = tmp; \
+      i--; \
+    } \
+ }
+
+
+#include <FL/Fl_Image.H>
+#include <string.h>
+#include <stdint.h>
+namespace { // anonymous namespace
+
+
+// Although we could use types from <stdint.h>, there is no assurance that int64_t is available.
+// Anyway the types below are reasonable expectations as:
+// 1) "int" is assumed by fltk to be at least 32 bit wide (otherwise fltk won't work)
+// 2) "long long" is the longest "standard" type available and it is usually 64 bit wide
+
+#ifndef uint32_t
+#define uint32_t unsigned int // use type at least 32 bit wide
+#endif
+#ifndef uint64_t
+#define uint64_t unsigned long long // use type at least 48 bit wide
+#endif
+
+
+uint32_t get_resample_scale(unsigned w1, unsigned w2) {
+  if(w2<w1) return w1;
+  return 2 * w2;
+}
+
+
+// resampling fron "unsigned char" to intermediate "unsigned int" array. The function returns the scalling factor.
+unsigned resample(unsigned w1, unsigned w2, const unsigned char * source, uint32_t * target, int source_stride, int target_stride, unsigned sets, int source_set_stride, int target_set_stride, int alpha_shift) {
+  if(w2<w1) {
+    for(unsigned j = 0; j<sets; j++) {
+      const unsigned char * sou= source + j * source_set_stride;
+      uint32_t * tar = target + j * target_set_stride;
+      if(alpha_shift)
+        RESAMPLE_DOWN(unsigned char, uint32_t, uint32_t, w1, w2, sou, tar, source_stride, target_stride, input *= src[alpha_shift]; , *tg = output;)
+      else
+        RESAMPLE_DOWN(unsigned char, uint32_t, uint32_t, w1, w2, sou, tar, source_stride, target_stride, , *tg = output;)
+    }
+    return w1;
+  }
+  for(unsigned j = 0; j<sets; j++) {
+    const unsigned char * sou= source + j * source_set_stride;
+    uint32_t * tar = target + j * target_set_stride;
+    if(alpha_shift)
+      RESAMPLE_UP(unsigned char, uint32_t, uint32_t, w1, w2, sou, tar, source_stride, target_stride, input *= src[alpha_shift]; , *tg = output;)
+    else
+      RESAMPLE_UP(unsigned char, uint32_t, uint32_t, w1, w2, sou, tar, source_stride, target_stride, , *tg = output;)
+  }
+  return 2 * w2;
+}
+
+
+
+// resampling from intermediate "unsigned" to final "unsigned char" with scalling
+void resample_scale(unsigned w1, unsigned w2, const uint32_t * source, unsigned char * target, int source_stride, int target_stride, unsigned sets, int source_set_stride, int target_set_stride, uint64_t scale, int alpha_shift) {
+  uint64_t sc_add = scale / 2;
+  uint64_t alpha;
+  if(w2<w1) {
+    for(unsigned j = 0; j<sets; j++) {
+      const uint32_t * sou= source + j * source_set_stride;
+      unsigned char * tar = target + j * target_set_stride;
+      if(alpha_shift)
+        RESAMPLE_DOWN(uint32_t , unsigned char, uint64_t, w1, w2, sou, tar, source_stride, target_stride, , alpha = scale * tg[alpha_shift]; if(!alpha) alpha = 1; *tg = (output + alpha/2)/alpha;  )
+      else
+        RESAMPLE_DOWN(uint32_t , unsigned char, uint64_t, w1, w2, sou, tar, source_stride, target_stride, , *tg = (output + sc_add)/scale;)
+    }
+    return;
+  }
+  for(unsigned j = 0; j<sets; j++) {
+    const uint32_t * sou= source + j * source_set_stride;
+    unsigned char * tar = target + j * target_set_stride;
+    if(alpha_shift)
+      RESAMPLE_UP(uint32_t, unsigned char, uint64_t, w1, w2, sou, tar, source_stride, target_stride, ,  alpha = scale * tg[alpha_shift];  if(!alpha) alpha = 1; *tg = (output + alpha/2)/alpha;)
+    else
+      RESAMPLE_UP(uint32_t, unsigned char, uint64_t, w1, w2, sou, tar, source_stride, target_stride, , *tg = (output + sc_add)/scale;)
+  }
+}
+
+
+
+// direct resampling  "unsigned char" to  "unsigned char" with scalling
+void resample_scale(unsigned w1, unsigned w2, const unsigned char * source, unsigned char * target, int source_stride, int target_stride, unsigned sets, int source_set_stride, int target_set_stride, uint32_t scale, int alpha_shift) {
+  uint32_t sc_add = scale / 2;
+  uint32_t alpha;
+  if(w2<w1) {
+    for(unsigned j = 0; j<sets; j++) {
+      const unsigned char * sou= source + j * source_set_stride;
+      unsigned char * tar = target + j * target_set_stride;
+      if(alpha_shift)
+        RESAMPLE_DOWN(unsigned char , unsigned char, uint32_t, w1, w2, sou, tar, source_stride, target_stride, , alpha = scale * tg[alpha_shift];  if(!alpha) alpha = 1;   *tg = (output + alpha/2)/alpha;)
+      else
+        RESAMPLE_DOWN(unsigned char , unsigned char, uint32_t, w1, w2, sou, tar, source_stride, target_stride, , *tg = (output + sc_add)/scale;)
+    }
+    return;
+  }
+  for(unsigned j = 0; j<sets; j++) {
+    const unsigned char * sou= source + j * source_set_stride;
+    unsigned char * tar = target + j * target_set_stride;
+    if(alpha_shift)
+      RESAMPLE_UP(unsigned char , unsigned char, uint32_t, w1, w2, sou, tar, source_stride, target_stride, , alpha = scale * tg[alpha_shift];  if(!alpha) alpha = 1;  *tg = (output + alpha/2)/alpha;)
+    else
+      RESAMPLE_UP(unsigned char , unsigned char, uint32_t, w1, w2, sou, tar, source_stride, target_stride, , *tg = (output + sc_add)/scale;)
+  }
+}
+
+}  // end of anonymous namespace
+
+
+// Finaly this is synthesis of the above functions to use optimal resizing approach.
+// Parameters:
+// w2, h2 - new (resampled) image size
+// source - pointer to original image data
+// w1, h1 - size of the original image
+// no_channels - number of adjacent channels (1 for mono images, 3 for RGB and 4 for RGBA or ARGB)
+// pixel_stride - pointer shift from pixel to pixel for the original image. If 0, "packed" data are assumed and pixel_stride equals to no_channels
+// row_stride - pointer shift between rows for the original image. If 0, row_stride = w1 * pixel_stride.
+// target - poinrer to array where resampled image data are stored. If 0, a new array is allocated with sufficient size.
+// target_pixel_stride - poinder shift from pixel to pixel for the resampled image. If 0, "packed" data are assumed and target_pixel_stride equals to no_channels
+// target_row_stride - pointer shift between rows for the resampled image. If 0, target_row_stride = w2 * target_pixel_stride.
+// The function returns the "target" parameter or pointer to newly allocated data (if target==0) - in such a case free this memory using operator delete[].
+
+unsigned char * resample(
+    int w2,
+    int h2,
+    const unsigned char * const source,
+    int w1,
+    int h1,
+    int no_channels,
+    int pixel_stride/* = 0*/,
+    int row_stride/* = 0*/,
+    unsigned char * target/* = 0*/,
+    int target_pixel_stride/* = 0*/,
+    int target_row_stride/* = 0*/,
+    bool alpha_premultiply/* = false*/) {
+
+  if(!pixel_stride) pixel_stride = no_channels;
+  if(!row_stride) row_stride = pixel_stride * w1;
+  const unsigned char * src = source;
+
+  int n_ch = no_channels - 1;
+  int alph_shift_multiplier = 0;
+  if(alpha_premultiply)
+    alph_shift_multiplier = 1;
+
+
+  if (pixel_stride <0) {
+    src += pixel_stride * (w1-1);
+    pixel_stride = -pixel_stride;
+  }
+  if (row_stride <0) {
+    src +=  row_stride * (h1-1);
+    row_stride = -row_stride;
+  }
+  if(!target_pixel_stride) target_pixel_stride = no_channels;
+  if(!target_row_stride) target_row_stride = target_pixel_stride * w2;
+  int size = target_row_stride *h2;
+  if(!target)
+    target = new unsigned char[size];
+
+  if(w1==w2 && h1==h2) { // direct copy
+    if(target_pixel_stride==pixel_stride) {
+      if(target_row_stride==row_stride)
+        memcpy(target, src, size * sizeof(unsigned char));
+      else for(int i = 0; i<h1; i++)
+          memcpy(target + i * target_row_stride , src + row_stride , row_stride * sizeof(unsigned char));
+    } else {
+      for(int j = 0; j<h1; j++) {
+        const unsigned char * s = src + j * row_stride;
+        unsigned char * t = target + j * target_row_stride;
+        int s_addition = pixel_stride - no_channels;
+        int t_addition = target_pixel_stride - no_channels;
+        for(int i = 0; i<w1; i++) {
+          for(int c = no_channels; c; c--)
+            *t++ = *s++;
+          s += s_addition;
+          t += t_addition;
+        }
+      }
+    }
+    return target;
+  }
+
+  if(w1==w2) { // no need for intermediate data, we resample only collumns
+    uint32_t scale = get_resample_scale(h1, h2);
+    resample_scale(h1, h2, source + n_ch, target + n_ch, row_stride, target_row_stride, w1, pixel_stride, target_pixel_stride, scale, 0);
+    for(int ch = 0; ch< n_ch; ch++)
+      resample_scale(h1, h2, source + ch, target + ch, row_stride, target_row_stride, w1, pixel_stride, target_pixel_stride, scale, alph_shift_multiplier * (n_ch - ch));
+    return target;
+  }
+  if(h1==h2) { // no need for intermediate data, we resample only rows
+    uint32_t scale = get_resample_scale(w1, w2);
+    resample_scale(w1, w2, source + n_ch, target + n_ch, pixel_stride, target_pixel_stride, h1, row_stride, target_row_stride, scale, 0);
+    for(int ch = 0; ch< n_ch; ch++)
+      resample_scale(w1, w2, source + ch, target + ch, pixel_stride, target_pixel_stride, h1, row_stride, target_row_stride, scale, alph_shift_multiplier * (n_ch - ch));
+    return target;
+  }
+
+  // General rescalling with intermediate uint32_t data array.
+  // First we try to find optimal approach to minimize intermediate data size:
+  int s = w2 * h1;
+  bool rows_first = 1;
+  {
+    int s2 = w1 * h2;
+    if(s2<s) {
+      s = s2;
+      rows_first = 0;
+    }
+  }
+
+  uint32_t * buffer = new uint32_t[s]; // intermediate buffer
+  uint64_t scale = ((uint64_t)(get_resample_scale(w1, w2))) * ((uint64_t)(get_resample_scale(h1, h2)));
+  if(rows_first) {
+    resample(w1, w2, source + n_ch, buffer, pixel_stride, 1, h1, row_stride, w2, 0);
+    resample_scale(h1, h2, buffer, target + n_ch, w2, target_row_stride, w2, 1, target_pixel_stride, scale, 0);
+    for(int ch = 0; ch< n_ch; ch++) {
+      resample(w1, w2, source + ch, buffer, pixel_stride, 1, h1, row_stride, w2, alph_shift_multiplier * (n_ch - ch));
+      resample_scale(h1, h2, buffer, target + ch, w2, target_row_stride, w2, 1, target_pixel_stride, scale, alph_shift_multiplier * (n_ch - ch));
+    }
+  } else {
+    resample(h1, h2, source + n_ch, buffer, row_stride, w1, w1, pixel_stride, 1, 0);
+    resample_scale(w1, w2, buffer, target + n_ch, 1, target_pixel_stride, h2, w1, target_row_stride, scale, 0);
+    for(int ch = 0; ch< n_ch; ch++) {
+      resample(h1, h2, source + ch, buffer, row_stride, w1, w1, pixel_stride, 1, alph_shift_multiplier * (n_ch - ch));
+      resample_scale(w1, w2, buffer, target + ch, 1, target_pixel_stride, h2, w1, target_row_stride, scale, alph_shift_multiplier * (n_ch - ch));
+    }
+  }
+  delete[] buffer;
+  return target;
+}
+
+// A wrapper for the resample method that can be called
+// conveniently from FLTK's image code
+
+/*static*/
+Fl_RGB_Image* Fl_RGB_Image::copy_bilinear(const Fl_RGB_Image& im, int W, int H) {
+  int row_stride = im.ld() ? im.ld() : im.d() * im.w() ;
+  unsigned char * array = resample(W, H, (const unsigned char*)*im.data(),
+    im.w(), im.h(), im.d(), im.d(), row_stride, 0, 0, 0, (im.d()==4));
+  Fl_RGB_Image * ni = new Fl_RGB_Image(array, W, H, im.d());
+  ni->alloc_array = 1;
+  return ni;
+}
Index: test/pixmap_browser.cxx
===================================================================
--- test/pixmap_browser.cxx	(revision 10301)
+++ test/pixmap_browser.cxx	(working copy)
@@ -20,6 +20,7 @@
 #include <FL/Fl_Box.H>
 #include <FL/Fl_Double_Window.H>
 #include <FL/Fl_Button.H>
+#include <FL/Fl_Check_Button.H>
 #include <FL/Fl_Shared_Image.H>
 #include <string.h>
 #include <errno.h>
@@ -29,8 +30,8 @@
 Fl_Box *b;
 Fl_Double_Window *w;
 Fl_Shared_Image *img;
+bool fit = false;
 
-
 static char name[1024];
 
 void load_file(const char *n) {
@@ -55,7 +56,7 @@
     b->redraw();
     return;
   }
-  if (img->w() > b->w() || img->h() > b->h()) {
+  if (fit || img->w() > b->w() || img->h() > b->h()) {
     Fl_Image *temp;
     if (img->w() > img->h()) temp = img->copy(b->w(), b->h() * img->h() / img->w());
     else temp = img->copy(b->w() * img->w() / img->h(), b->h());
@@ -67,11 +68,12 @@
   b->labelsize(14);
   b->labelcolor(FL_FOREGROUND_COLOR);
   b->image(img);
-  b->redraw();
+//  b->redraw();
+  w->redraw();
 }
 
 void file_cb(const char *n) {
-  if (!strcmp(name,n)) return;
+//  if (!strcmp(name,n)) return;
   load_file(n);
   strcpy(name,n);
   w->label(name);
@@ -84,6 +86,19 @@
   fl_file_chooser_callback(0);
 }
 
+void scaling_cb(Fl_Widget *o,void *) {
+  Fl_Check_Button *b = (Fl_Check_Button *)o;
+  int bilinear = b->value();
+  Fl_RGB_Image::scaling(bilinear ? FL_SCALING_BILINEAR : FL_SCALING_NEAREST);
+  file_cb(name);
+}
+
+void fit_cb(Fl_Widget *o,void *) {
+  Fl_Check_Button *b = (Fl_Check_Button *)o;
+  fit = b->value();
+  file_cb(name);
+}
+
 int dvisual = 0;
 int arg(int, char **argv, int &i) {
   if (argv[i][1] == '8') {dvisual = 1; i++; return 1;}
@@ -96,6 +111,7 @@
   fl_register_images();
 
   Fl::args(argc,argv,i,arg);
+  Fl_RGB_Image::scaling(FL_SCALING_NEAREST);
 
   Fl_Double_Window window(400,435); ::w = &window;
   Fl_Box b(10,45,380,380); ::b = &b;
@@ -103,6 +119,12 @@
   b.align(FL_ALIGN_INSIDE|FL_ALIGN_CENTER);
   Fl_Button button(150,5,100,30,"load");
   button.callback(button_cb);
+  Fl_Check_Button bi(15,5,100,30,"bilinear\nscaling");
+  bi.labelsize(12);
+  bi.callback(scaling_cb);
+  Fl_Check_Button fit(310,5,100,30,"scale to fit");
+  fit.labelsize(12);
+  fit.callback(fit_cb);
   if (!dvisual) Fl::visual(FL_RGB);
   if (argv[1]) load_file(argv[1]);
   window.resizable(window);
